/*
 * BSD 2 clause license
 *
 * Author: Wojciech A. Koszek <wkoszek@FreeBSD.org>
 * 
 * This file implements a functions to handle Xilinx .bit file
 * header format. Layout was taken form:
 *
 * http://www.fpga-faq.com/FAQ_Pages/0026_Tell_me_about_bit_files.htm
 *
 * Description contains 2 "Field 4" sections -- and there is 7, not 6
 * fields. It's changed within this file. 
 */
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <netinet/in.h>

#include <assert.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include "xbf.h"

#define ASSERT	assert

static int
_xbf_setup(struct xbf *xbf)
{
	uint32_t u32;
	uint16_t u16;
	uint8_t u8;
	char *ptr;
	char *endptr;

	ASSERT(xbf != NULL);

	/*
	 * Helper functions that let me to access a memory with
	 * different granularity easily from a byte pointer.
	 */
#define U32(ptr)			\
	(htonl(*(uint32_t *)ptr))
#define U16(ptr)			\
	(htons(*(uint16_t *)ptr))
#define U8(ptr)				\
	(*(char *)(ptr))

	ptr = xbf->_xbf_mem;
	endptr = (char *)xbf->_xbf_mem + xbf->_xbf_memsize;
#define LEFT()				\
	((unsigned)(endptr - ptr))

#define WHDR "Wrong header format! "

	/*
	 * Field 1:
	 * 2 bytes          length 0x0009           (big endian) 
	 * 9 bytes          some sort of header
	 */
	u16 = U16(ptr);
	if (u16 != 9)
		return (xbf_erri(xbf, WHDR "1st field length should be 9, "
		    "but is %d", u16));
	ptr += 2;
	ptr += u16;

	/*
	 * Field 2
	 * 2 bytes          length 0x0001
	 * 1 byte           key 0x61                (The letter "a")
	 */
	u16 = U16(ptr);
	ptr += 2;
	u8 = U8(ptr);
	ptr += 1;
	if (u8 != 'a')
		return (xbf_erri(xbf, WHDR "Magic 'a' missing"));

	/*
	 * Field 3 
	 * 2 bytes          length 0x000a           (value depends on file name length) 
	 * 10 bytes         string design name "xform.ncd" (including a trailing 0x00) 
	 */
	u16 = U16(ptr);
	ptr += 2;
	xbf->xbf_ncdname = ptr;
	if (u16 > LEFT())
		return (xbf_erri(xbf, WHDR "NCD filename seems to be to long"));
	ptr += u16;

	/*
	 * Field 4 
	 * 1 byte           key 0x62                (The letter "b") 
	 * 2 bytes          length 0x000c           (value depends on part name length) 
	 * 12 bytes         string part name "v1000efg860" (including a  trailing 0x00)
	 */
	u8 = U8(ptr);
	ptr += 1;
	u16 = U16(ptr);
	ptr += 2;
	xbf->xbf_partname = ptr;
	if (u16 > LEFT())
		return (xbf_erri(xbf, WHDR "Part name to long"));
	ptr += u16;

	/*
	 * Field 5
	 * 1 byte           key 0x63                (The letter "c") 
	 * 2 bytes          length 0x000b 
	 * 11 bytes         string date "2001/08/10"  (including a trailing 0x00)
	 */
	u8 = U8(ptr);
	ptr += 1;
	if (u8 != 'c')
		return (xbf_erri(xbf, WHDR "Magic 'c' missing"));
	u16 = U16(ptr);
	if (u16 != 11)
		return (xbf_erri(xbf, WHDR "5th field length should be 11, "
		    "but is %d", u16));
	ptr += 2;
	xbf->xbf_date = ptr;
	ptr += u16;

	/*
	 * Field 6
	 * 1 byte           key 0x64                (The letter "d") 
	 * 2 bytes          length 0x0009 
	 * 9 bytes          string time "06:55:04"    (including a trailing 0x00)
	 */
	u8 = U8(ptr);
	ptr += 1;
	if (u8 != 'd')
		return (xbf_erri(xbf, WHDR "Magic 'd' missing"));
	u16 = U16(ptr);
	if (u16 != 9)
		return (xbf_erri(xbf, WHDR "6th field length should be 9, "
		    "but is %d", u16));
	ptr += 2;
	xbf->xbf_time = ptr;
	ptr += u16;

	/*
	 * Field 7 
	 * 1 byte           key 0x65                 (The letter "e") 
	 * 4 bytes          length 0x000c9090        (value depends on device type,
	 * and maybe design details) 
	 */
	u8 = U8(ptr);
	ptr += 1;
	if (u8 != 'e')
		return (xbf_erri(xbf, WHDR "Magic 'e' missing"));
	u32 = U32(ptr);
	if (u32 > LEFT())
		return (xbf_erri(xbf, WHDR "Part name filename seems to be to long"));
	ptr += 4;
	xbf->xbf_bitlen = u32;
	xbf->xbf_bitptr = ptr;
	return (0);
}

static void
xbf_err_fmt(struct xbf *xbf, const char *fmt, va_list va)
{
	int len;
	int esz;

	ASSERT(xbf != NULL);
	ASSERT(fmt != NULL);

	memset(xbf->_xbf_errmsg, 0, sizeof(xbf->_xbf_errmsg));
	esz = sizeof(xbf->_xbf_errmsg) - 1;
	len = vsnprintf(xbf->_xbf_errmsg, esz, fmt, va);
}

struct xbf *
xbf_err(struct xbf *xbf, const char *fmt, ...)
{
	va_list va;

	va_start(va, fmt);
	xbf_err_fmt(xbf, fmt, va);
	va_end(va);
	return (xbf);
}

int
xbf_erri(struct xbf *xbf, const char *fmt, ...)
{
	va_list va;

	va_start(va, fmt);
	xbf_err_fmt(xbf, fmt, va);
	va_end(va);
	return (-1);
}

const char *
xbf_errmsg(struct xbf *xbf)
{

	ASSERT(xbf != NULL);
	return (xbf->_xbf_errmsg);
}

int
xbf_opened(struct xbf *xbf)
{

	ASSERT(xbf != NULL);
	return (xbf->_xbf_errmsg[0] == '\0');
}

struct xbf *
xbf_open(const char *fname)
{
	struct xbf *xbf;
	struct stat st;
	int fd;
	int error;

	xbf = calloc(1, sizeof(*xbf));
	ASSERT(xbf != NULL);

	xbf->xbf_fname = fname;
	fd = open(fname, O_RDONLY);
	if (fd == -1)
		return (xbf_err(xbf, "Couldn't open file '%s'", fname));
	memset(&st, 0, sizeof(st));
	error = fstat(fd, &st);
	if (error == -1)
		return (xbf_err(xbf, "Couldn't check file '%s' information",
		    fname));
	if (st.st_size <= XBF_HDR_SIZE)
		return (xbf_err(xbf, "File '%s' doesn't contain valid data",
		    fname));
	xbf->_xbf_mem = mmap(NULL, st.st_size, PROT_READ|PROT_WRITE,
	    MAP_PRIVATE, fd, 0);
	if (xbf->_xbf_mem == MAP_FAILED)
		return (xbf_err(xbf, "Couldn't map file '%s' to memory", fname));
	xbf->_xbf_memsize = st.st_size;
	(void)_xbf_setup(xbf);
	return (xbf);
}

int
xbf_close(struct xbf *xbf)
{
	int error;

	ASSERT(xbf != NULL);
	ASSERT(xbf->_xbf_mem != NULL);
	error = munmap(xbf->_xbf_mem, xbf->_xbf_memsize);
	ASSERT(error == 0);

	return (error);
}

const char *
xbf_get_time(struct xbf *xbf)
{

	return (xbf->xbf_time);
}

const char *
xbf_get_date(struct xbf *xbf)
{

	return (xbf->xbf_date);
}

const char *
xbf_get_partname(struct xbf *xbf)
{

	return (xbf->xbf_partname);
}

const char *
xbf_get_ncdname(struct xbf *xbf)
{

	return (xbf->xbf_ncdname);
}

const char *
xbf_get_fname(struct xbf *xbf)
{

	return (xbf->xbf_fname);
}

size_t
xbf_get_bitlen(struct xbf *xbf)
{

	return (xbf->xbf_bitlen);
}

const unsigned char *
xbf_get_bitptr(struct xbf *xbf)
{

	return (xbf->xbf_bitptr);
}

void
xbf_print_fp(FILE *fp, struct xbf *xp)
{

	ASSERT(fp != NULL);
	ASSERT(xp != NULL);

	(void)fprintf(fp, "NCD filename: %s\n", xbf_get_ncdname(xp));
	(void)fprintf(fp, "   Part name: %s\n", xbf_get_partname(xp));
	(void)fprintf(fp, "        Date: %s\n", xbf_get_date(xp));
	(void)fprintf(fp, "        Time: %s\n", xbf_get_time(xp));
	(void)fprintf(fp, "Image lenght: %d\n", xbf_get_bitlen(xp));
}

void
xbf_print(struct xbf *xbf)
{

	xbf_print_fp(stdout, xbf);
}

#ifdef XBF_TEST_PROG

int
main(int argc, char **argv)
{
	struct xbf *xbf;
	int flag_v = 0;
	char *fname = NULL;
	int o = -1;
	char *prog = NULL;

	prog = argv[0];
	while ((o = getopt(argc, argv, "v")) != -1)
		switch (o) {
		case 'v':
			flag_v++;
			break;
		}

	argc -= optind;
	argv += optind;

	if (argc == 0)
		errx(EX_USAGE, "usage: %s filename ...", prog);

	fname = argv[0];

	xbf = xbf_open(fname);
	if (!xbf_opened(xbf))
		errx(EXIT_FAILURE, "%s", xbf_errmsg(xbf));
	xbf_print(xbf);
	xbf_close(xbf);

	exit(EXIT_SUCCESS);
}
#endif /* XBF_TEST_PROG */
