/*
 * NetFPGA v2 experimentation sandbox for FreeBSD
 *
 * BSD 2nd clause license
 * Author: Wojciech A. Koszek <wkoszek@FreeBSD.org>
 *
 * TODO:
 *
 * Design small and simple tool that will be able to do NetFPGA host
 * stuff without extreme code duplication. It should be basically
 * something like:
 *
 * netfpga image read <file>
 * netfpga image write <file>
 * netfpga image info
 * 
 * netfpga reg read [-f file] <addr>
 * netfpga reg write <add
 * netfpga reg list
 * 
 * netfpga server
 * netfpga client
 *
 * --- number of plugins might exist ----
 * - switch
 * - nic
 * - router
 *
 * Currently, native tools have 5 lines of functionality in 100 lines
 * of argument handling and it's still impossible to request specific
 * registers without knowing their offsets.
 *
 * This tool is supposed to be used with NetFPGA driver for FreeBSD kernel
 * that is being developed as a part of this work.
 * 
 * $Id$
 */
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/utsname.h>

#include <netinet/in.h>

#include <assert.h>
#include <ctype.h>
#include <dlfcn.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include "../../include/nf2.h"
#include "../../include/nf2_common.h"
#include "../../include/netfpga_freebsd.h"
#include "../libxbf/xbf.h"

#include "netfpga.h"

#define NF_WR32	netfpga_wr32
#define NF_RD32	netfpga_rd32

int netfpga_debug = 0;
int netfpga_tbd = 0;

int
netfpga_has_error(struct netfpga *nf)
{

	return (nf->__nf_errmsg[0] != '\0');
}

const char *
netfpga_strerror(struct netfpga *nf)
{

	return (nf->__nf_errmsg);
}

void
_netfpga_err_fmt(struct netfpga *nf, const char *func, int lineno,
    const char *fmt, va_list va)
{

	NETFPGA_ASSERT(nf);
	ASSERT(func != NULL);
	ASSERT(fmt != NULL);

	/*
	 * Fill the error message only if we don't have a message yet.
	 * Do nothing otherwise. This lets us to have error handling
	 * that propagates from embedded functions.
	 */
	if (netfpga_has_error(nf)) {
		printf("error already present\n");
		return;
	}

	memset(nf->__nf_errmsg_src, 0, sizeof(nf->__nf_errmsg_src));
	memset(nf->__nf_errmsg, 0, sizeof(nf->__nf_errmsg));
	snprintf(nf->__nf_errmsg_src, sizeof(nf->__nf_errmsg_src) - 1,
	    "%s(%d): ", func, lineno);
	vsnprintf(nf->__nf_errmsg, sizeof(nf->__nf_errmsg), fmt, va);
}

struct netfpga *
_netfpga_err(struct netfpga *nf, const char *func, int lineno, const char *fmt, ...)
{
	va_list va;

	va_start(va, fmt);
	_netfpga_err_fmt(nf, func, lineno, fmt, va);
	va_end(va);

	return (nf);
}

int
_netfpga_erri(struct netfpga *nf, const char *func, int lineno, const char *fmt, ...)
{
	va_list va;

	va_start(va, fmt);
	_netfpga_err_fmt(nf, func, lineno, fmt, va);
	va_end(va);

	return (-lineno);
}

int
netfpga_start(struct netfpga *nf)
{
	char iomod_fname[MAXPATHLEN];
	struct utsname ut;
	unsigned i;
	char *tmp;
	int error = 0;
	netfpga_open_t *nfopen;
	void *ctx;
	char *modname;

	ASSERT(nf != NULL);
	tmp = NULL;
	_netfpga_start(nf);
	NETFPGA_INIT(nf);

	memset(&ut, 0, sizeof(ut));
	error = uname(&ut);
	if (error != 0)
		return (netfpga_erri(nf, "Couldn't get system type (uname(2)"
		    " failed)!"));
	tmp = calloc(1, strlen(ut.sysname) + 1);
	ASSERT(tmp != NULL);
	for (i = 0; i < strlen(ut.sysname); i++)
		tmp[i] = tolower(ut.sysname[i]);
	snprintf(iomod_fname, sizeof(iomod_fname), "%s/netfpga_%s.so",
	    NETFPGA_PATH_LIB, tmp);
	free(tmp);
	nf->__nf_dll = dlopen(iomod_fname, RTLD_NOW);
	if (nf->__nf_dll == NULL)
		return (netfpga_erri(nf, "Can't get '%s' module to load or module"
		    " has been compiled improperly", iomod_fname));
	modname = dlsym(nf->__nf_dll, "__start_netfpga");
	if (modname == NULL)
		return (netfpga_erri(nf, "Unable to fetch module name from %s",
		    iomod_fname));
	nf->__nf_mod = dlsym(nf->__nf_dll, modname);
	if (nf->__nf_mod == NULL)
		return (netfpga_erri(nf, "Couldn't find valid modules in %s",
		    iomod_fname));
	nfopen = nf->__nf_mod->nf_open;
	if (nfopen == NULL)
		return (netfpga_erri(nf, "There is no 'open' method in a module"));
	ctx = nfopen(nf);
	if (ctx == NULL)
		return (netfpga_erri(nf, "Method 'open' failed"));
	nf->__nf_mod_ctx = ctx;
	return (error);
}

int
netfpga_stop(struct netfpga *nf)
{
	int error;
	netfpga_close_t *nfclose;

	NETFPGA_ASSERT(nf);
	if (nf->__nf_dll == NULL)
		return (netfpga_erri(nf, "No module opened -- internal error"));
	if (nf->__nf_mod == NULL)
		return (netfpga_erri(nf, "No module loaded -- internal error"));
	ASSERT(nf->__nf_dll != NULL);
	ASSERT(nf->__nf_mod != NULL);
	nfclose = nf->__nf_mod->nf_close;
	if (nfclose == NULL)
		return (netfpga_erri(nf, "There is no 'close' method' in a module"));
	error = nfclose(nf, nf->__nf_mod_ctx);
	if (error != 0)
		return (-1);
	error = dlclose(nf->__nf_dll);
	if (error != 0)
		return (netfpga_erri(nf, "Can't close handler for module"));
	return (error);
}

int
netfpga_read(struct netfpga *nf, uint32_t reg, void *buf, size_t buf_len)
{

	NETFPGA_ASSERT(nf);
	ASSERT(buf != NULL);
	ASSERT(nf->__nf_mod != NULL);
	ASSERT(nf->__nf_mod->nf_read != NULL);
	ASSERT(reg % 4 == 0 && "must be 4 aligned");
	ASSERT(buf_len % 4 == 0 && "must be 4 aligned");
	return (nf->__nf_mod->nf_read(nf, nf->__nf_mod_ctx, reg, buf, buf_len));
}

int
netfpga_write(struct netfpga *nf, uint32_t reg, void *buf, size_t buf_len)
{

	NETFPGA_ASSERT(nf);
	ASSERT(buf != NULL);
	ASSERT(nf->__nf_mod != NULL);
	ASSERT(nf->__nf_mod->nf_read != NULL);
	ASSERT(reg % 4 == 0 && "must be 4 aligned");
	ASSERT(buf_len % 4 == 0 && "must be 4 aligned");
	return (nf->__nf_mod->nf_write(nf, nf->__nf_mod_ctx, reg, buf, buf_len));
}

uint32_t
netfpga_rd32(struct netfpga *nf, uint32_t reg)
{
	uint32_t u32;
	int ret;

	ret = netfpga_read(nf, reg, &u32, sizeof(u32));
	ASSERT(ret == sizeof(u32));
	return (u32);
}

void
netfpga_wr32(struct netfpga *nf, uint32_t reg, uint32_t value)
{
	int ret;

	ret = netfpga_write(nf, reg, &value, sizeof(value));
	ASSERT(ret == sizeof(value));
}

void
netfpga_image_devname(struct netfpga *nf, char **dev_name)
{
	uint32_t *u32;
	char _dn[NF2_DEVICE_STR_LEN];
	int i;

	ASSERT(dev_name != NULL);
	memset(_dn, 0, sizeof(_dn));

	u32 = (uint32_t *)_dn;
	for (i = 0; i < NF2_DEVICE_STR_LEN / 4; i += 4, u32++)
		*u32 = htonl(netfpga_rd32(nf, NF2_DEVICE_STR + i));
	*dev_name = strdup(_dn);
	ASSERT(*dev_name != NULL);
}

static int
netfpga_image_valid(struct netfpga *nf, struct xbf *xbf)
{
	uint32_t version;
	unsigned exblen;

	NETFPGA_ASSERT(nf);
	ASSERT(xbf != NULL);

	/* Board version */
	version = NF2_GET_VERSION(NF_RD32(nf, CPCI_REG_ID));
	DEBUG("Board version: %d\n", version);

	/* 
	 * Before we touch anything, check if we got the correct
	 * bitstream file.
	 */
	if (version == 1)
		exblen = VIRTEX_BIN_SIZE_V2_0;
	else if (version == 2)
		exblen = VIRTEX_BIN_SIZE_V2_1;
	else
		return (netfpga_erri(nf, "Version %d of the board unsupported",
		    version));
	if (exblen != xbf_get_bitlen(xbf))
		return (netfpga_erri(nf, "Expected that bit stream length "
		    "of '%s' to be %d, but it's %d", xbf_get_fname(xbf),
		    exblen, xbf_get_bitlen(xbf)));
	return (1);
}

static int
netfpga_prog_reset(struct netfpga *nf)
{
	uint32_t status;
	uint32_t errreg;

	NETFPGA_ASSERT(nf);

	/* Reset the programming interface */
	NF_WR32(nf, CPCI_REG_PROG_CTRL, RESET_CPCI);

	/* Clear error register */
	NF_WR32(nf, CPCI_ERROR, 0);

	/* Sleep a bit */
	usleep(100);

	/* 
	 * Read programming status: Check that DONE bit (bit 8) is zero
	 * and FIFO (bit 1) is empty (1).
	 */
	status = NF_RD32(nf, CPCI_REG_PROG_STATUS);
	if ((status & (PROG_DONE | PROG_FIFO_EMPTY)) != PROG_FIFO_EMPTY) {
		return (netfpga_erri(nf, "After programming interface reset,"
			"FIFO isn't empty(%s %s)",
			status & PROG_DONE ? "PROG_DONE " : "",
			status & PROG_FIFO_EMPTY ? " PROG_FIFO_EMPTY" : ""
		    )
		);
	}

	/* XXWKOSZEK: Try to find error bits and print something
	 * sensible ONLY if something goes wrong.
	 * Error registers. nf2_download consistency
	 */
	errreg = NF_RD32(nf, CPCI_REG_ERROR);
	fprintf(stderr, "Error Registers: %x\n", errreg);

	/* XXWKOSZEK: This doesn't make much sense */
	fprintf(stderr, "Good, after resetting programming interface the FIFO is empty\n");

	/* Sleep a bit */
	usleep(1000);
	status = NF_RD32(nf, CPCI_REG_PROG_STATUS);
	ASSERT(status & (PROG_INIT | PROG_FIFO_EMPTY));
	return (1);
}

static int
netfpga_prog_done(struct netfpga *nf)
{
	uint32_t status;
	int tries;

	/* 
	 * Check, if everything is fine once we reach this stage and
	 * wait for some time if we aren't done yet.
	 */
	tries = 5;
	do {
		status = NF_RD32(nf, CPCI_REG_PROG_STATUS);
		if (status & PROG_DONE) {
			fprintf(stderr, "DONE went high - chip has been "
			    "successfully programmed.\n");
			break;
		}
		sleep(1);
		tries--;
	} while (tries > 0);

	if (tries == 0) {
		printf("status = %#x\n", status);
		if (status & PROG_INIT)
			fprintf(stderr, "INIT went high - appears to be a programming error.\n");
		if (!(status & PROG_DONE))
			fprintf(stderr, "DONE has not gone high - looks like an error\n");
		return (netfpga_erri(nf, "Error while programming NetFPGA Virtex chip"));
	}
	return (1);
}

static int
netfpga_image_write_start(struct netfpga *nf, struct xbf *xbf)
{
	const uint32_t *prog_word;
	uint32_t status;
	int nwrites = 0;
	size_t bytes_written = 0;
	int tries;

	NETFPGA_ASSERT(nf);
	ASSERT(xbf != 0);

	(void)tries;
	(void)status;

	bytes_written = 0;
	prog_word = (const uint32_t *)xbf_get_bitptr(xbf);
	printf("Expected to write = %d\n", xbf_get_bitlen(xbf));

	nwrites = xbf_get_bitlen(xbf) / 4;
	printf("Expected loop itarations: %d\n", nwrites);

	while (nwrites) {
		NF_WR32(nf, CPCI_REG_PROG_DATA, *prog_word);
		prog_word++;
		bytes_written += 4;
		nwrites--;
		if (nwrites != 0 && nwrites % 8 == 0) {
			printf(".");
			tries = 5;
			do {
				status = NF_RD32(nf, CPCI_REG_PROG_STATUS);
				if (status & PROG_DONE)
					break;
				if (!(status & PROG_FIFO_EMPTY))
					sleep(1);
				tries--;
			} while (tries > 0);
			ASSERT(tries != 0);

//			printf("%d/%d bytes\n", bytes_written, xbf_get_bitlen(xbf));
#if 0
			tries = 3;
			do {
				status = NF_RD32(nf, CPCI_REG_PROG_STATUS);
				if (!(status & PROG_DONE)) {
					usleep(200000);
				}
				status = NF_RD32(nf, CPCI_REG_PROG_STATUS);
				if (!(status & PROG_FIFO_EMPTY))
					usleep(200000);
				break;
			} while (tries--);
			//status = NF_RD32(nf, CPCI_REG_PROG_STATUS);
			//ASSERT((status & PROG_DONE) && (status & PROG_FIFO_EMPTY));
#endif
		}
	}
	printf("Bytes_written = %d\n", bytes_written);
	return (bytes_written);
}

static int
netfpga_image_write(struct netfpga *nf, const char *fname)
{
	struct xbf *xbf;

	NETFPGA_ASSERT(nf);

	xbf = xbf_open(fname);
	if (!xbf_opened(xbf))
		return (netfpga_erri(nf, "Problem with the image file: %s",
		    xbf_errmsg(xbf)));
	if (!netfpga_image_valid(nf, xbf))
		return (netfpga_erri(nf, "Invalid image for this NetFPGA"
		    " card"));
	if (!netfpga_prog_reset(nf))
		return (netfpga_erri(nf, "Couldn't get CPCI programmer to "
		    "reset"));
	if (!netfpga_image_write_start(nf, xbf))
		return (netfpga_erri(nf, "Couldn't program a device"));
	if (!netfpga_prog_done(nf)) {
		xbf_close(xbf);
		return (netfpga_erri(nf, "Error occured while card"
		    "programming"));
	}
	xbf_close(xbf);
	return (0);
}

#if 0
netfpga_image_write();
netfpga_image_write_cpci();
netfpga_image_devname();
netfpga_reg_readbyname();
netfpga_reg_writebyname();
netfpga_reg_list();
netfpga_reset();
#endif

#ifdef NETFPGA_PROG
static void
usage(const char *prog)
{

	printf("USAGE:\n");
	printf("%s: Look at source code\n", prog);
	exit(EX_USAGE);
}

/*
 * Sample program that will be used until I analyze the requirements for
 * the userland tools. Please don't get used to the command line
 * arguments.
 *
 * This stuff is only for debugging.
 */
int
main(int argc, char **argv)
{
	struct netfpga nf;
	const char *prog = NULL;
	int error = 0;
	int o = -1;
	char *arg_iface = NULL;
	int flag_verbose = 0;
	char *imgname = NULL;

	setbuf(stdout, NULL);
	setbuf(stderr, NULL);

	prog = argv[0];
	while ((o = getopt(argc, argv, "hi:v")) != -1)
		switch (o) {
		case 'h':
			usage(prog);
			break;
		case 'i':
			arg_iface = optarg;
			break;
		case 'v':
			flag_verbose++;
			break;
		}

	argc += optind;
	argv += optind;

	memset(&nf, 0, sizeof(nf));
	nf.nf_iface = arg_iface;
	nf.nf_verbose = flag_verbose;
	error = netfpga_start(&nf);
	if (error != 0)
		err(EXIT_FAILURE, "%s", netfpga_strerror(&nf));
if (0) {
	netfpga_image_devname(&nf, &imgname);
	printf("Image name: %s\n", imgname);
	free(imgname);
}
if (1) {
	DEBUG("B");
	netfpga_image_write(&nf, "/home/wkoszek/cpci_reprogrammer.bit");
	DEBUG("A");
}
	error = netfpga_stop(&nf);
	if (error != 0)
		err(EXIT_FAILURE, "%s", netfpga_strerror(&nf));
	exit(EXIT_SUCCESS);
}
#endif
