#include <sys/types.h>
#include <sys/ioccom.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#include "../../include/nf2_common.h"
#include "../../include/nf2.h"
#include "../../include/netfpga_freebsd.h"

#include "netfpga.h"

#define NETFPGA_FREEBSD_DEVNAME	"/dev/netfpga0"

netfpga_open_t nf2_freebsd_open;
netfpga_close_t nf2_freebsd_close;
netfpga_read_t nf2_freebsd_read;
netfpga_write_t nf2_freebsd_write;

struct netfpga_softc {
	int fd;
};

void *
nf2_freebsd_open(struct netfpga *nf)
{
	struct netfpga_softc *sc;
	const char *dev_name;

	TBD();
	sc = calloc(1, sizeof(*sc));
	ASSERT(sc != NULL);
	dev_name = NETFPGA_FREEBSD_DEVNAME;
	if (nf->nf_iface != NULL)
		dev_name = nf->nf_iface;
	sc->fd = open(dev_name, O_RDWR);
	if (sc->fd == -1) {
		(void)netfpga_erri(nf, "Couldn't open device %s", dev_name);
		free(sc);
		return (NULL);
	}
	return (sc);
}

int
nf2_freebsd_close(struct netfpga *nf, void *ctx)
{
	struct netfpga_softc *sc;
	int error;

	TBD();
	ASSERT(ctx != NULL);
	sc = ctx;
	error = -1;

	if (sc->fd == -1) {
		(void)netfpga_erri(nf, "No valid descriptor found--internal error");
		goto errout;
	}
	error = close(sc->fd);
	if (error == -1) {
		(void)netfpga_erri(nf, "Couldn't close device descriptor");
		goto errout;
	}
errout:
	free(sc);
	return (error);
}

int
nf2_freebsd_read(struct netfpga *nf, void *ctx, uint32_t reg, void *buf, size_t buf_len)
{
	struct netfpga_softc *sc;
	struct netfpga_req req;
	int error = 0;
	uint32_t *u32;
	unsigned int i;
	int bytes_read;

	TBD();
	ASSERT(nf != NULL);
	ASSERT(buf != NULL);
	ASSERT(ctx != NULL);
	sc = ctx;
	memset(&req, 0, sizeof(req));

	if (buf_len % 4 != 0)
		return (netfpga_erri(nf, "Buffer length must be a multiple of 4 bytes"));
	
	bytes_read = 0;
	for (u32 = buf, i = 0; i < buf_len / 4; i += 4) {
		req.offset = reg + i;
		error = ioctl(sc->fd, SIOCREGREAD, &req);
		ASSERT(error == 0);
		*u32 = req.value;
		DEBUG("READ, req.value=%llx(%lld)\n", req.value, req.value);
		bytes_read += 4;
	}
	return (bytes_read);
}

int
nf2_freebsd_write(struct netfpga *nf, void *ctx, uint32_t reg, void *buf, size_t buf_len)
{
	struct netfpga_softc *sc;
	struct netfpga_req req;
	int error = 0;
	uint32_t *u32;
	unsigned int i;
	int written;

	TBD();
	ASSERT(nf != NULL);
	ASSERT(buf != NULL);
	ASSERT(ctx != NULL);
	sc = ctx;
	memset(&req, 0, sizeof(req));

	if (buf_len % 4 != 0)
		return (netfpga_erri(nf, "Buffer length must be a multiple of 4 bytes"));
	
	written = 0;
	for (u32 = buf, i = 0; i < buf_len / 4; i += 4) {
		req.offset = reg + i;
		req.value = *u32++;
		error = ioctl(sc->fd, SIOCREGWRITE, &req);
		ASSERT(error == 0);
		DEBUG("WRITE, req.offset=%llx(%lld), req.value=%llx(%lld)\n",
		    req.offset, req.offset,
		    req.value, req.value);
		written += 4;
	}
	return (written);
}

struct netfpga_module nf2_freebsd = {
	.nf_version =	0,
	.nf_open =	nf2_freebsd_open,
	.nf_close = 	nf2_freebsd_close,
	.nf_read =	nf2_freebsd_read,
	.nf_write =	nf2_freebsd_write,
};
NETFPGA_MODULE(nf2_freebsd);
