/*
 * Author: Wojciech A. Koszek <wkoszek@FreeBSD.org>
 *
 * $Id$
 */
#include <sys/param.h>
#include <sys/bus.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/ioccom.h>

#include <sys/resource.h>
#include <machine/bus.h>
#include <sys/rman.h>


#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>

#include "../../include/nf2.h"
#include "../../include/nf2_common.h"
#include "../../include/netfpga_freebsd.h"

/*--------------------------------------------------------------------------*/
static int nf2_debug = 1;

#define TBD() do {						\
		printf("%s(%d): [Still to be implemented]\n",	\
		    __func__, __LINE__);			\
} while (0)
#define NF2_DEBUG_LV(lv, ...) do {				\
		if (nf2_debug >= (lv)) {			\
			printf("%s(%d): ", __func__, __LINE__);	\
			printf(__VA_ARGS__);			\
			printf("\n");				\
		}						\
} while (0)
#define NF2_DEBUG(...)	NF2_DEBUG_LV((1), __VA_ARGS__)
#define NF2_DEBUG3(...)	NF2_DEBUG_LV((3), __VA_ARGS__)
#define NF2_ASSERT(x) do {					\
	if (!(x)) {						\
		printf("___.ooo.__|__|__.ooo._---------____\n");\
		printf("%s(%d): Assertion '%s' failed\n",	\
		    __func__, __LINE__, (#x));			\
		printf("___.ooo.__|__|__.ooo._---------____\n");\
	}							\
} while (0)

/*--------------------------------------------------------------------------*/

struct nf2_softc {
	unsigned nf2_softc_magic;
#define	NF2_SOFTC_MAGIC		0x53e978b3 /* from /dev/random */
#define	NF2_SOFTC_INIT(n)	((n)->nf2_softc_magic = NF2_SOFTC_MAGIC)
#define	NF2_SOFTC_ASSERT(n)	NF2_ASSERT((n)->nf2_softc_magic == NF2_SOFTC_MAGIC)
	int lineno;
	const char *func;

	device_t dev;
	int revid;
	struct cdev *cdev;
	struct resource *mem;
	bus_space_tag_t mem_tag;
	bus_space_handle_t mem_handle;
	struct resource *ioport;
	struct resource *irq;
	void *intrhand;

	bus_dma_tag_t	parent_tag;
	bus_dma_tag_t	rx_tag;

	void *mem_kva;
	int need_reset;
};

/* Helper macros to make memory accesses faster to type */
#define WR32(sc, offset, v)							\
	bus_space_write_4((sc)->mem_tag, (sc)->mem_handle, (offset), (v))
#define RD32(sc, offset)							\
	bus_space_read_4((sc)->mem_tag, (sc)->mem_handle, (offset))

/*--------------------------------------------------------------------------*/

#define VENDOR_ID_STANFORD	0xFEED
#define DEVICE_ID_STANFORD_NF2	0x0001

static int nf2_probe(device_t);
static int nf2_attach(device_t);
static int nf2_detach(device_t);

static device_method_t nf2_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,		nf2_probe),
	DEVMETHOD(device_attach,	nf2_attach),
	DEVMETHOD(device_detach, 	nf2_detach),

	{ NULL, NULL }
};

static driver_t nf2_driver = {
	"nf2",
	nf2_methods,
	sizeof(struct nf2_softc)
};

static devclass_t nf2_devclass;

DRIVER_MODULE(nf2, pci, nf2_driver, nf2_devclass, 0, 0);

static d_ioctl_t nf2_ioctl;
static d_open_t nf2_open;
static d_close_t nf2_close;

static struct cdevsw nf2_cdevsw = {
	.d_version =	D_VERSION,
	.d_flags =	D_NEEDGIANT,
	.d_ioctl =	nf2_ioctl,
	.d_open =	nf2_open,
	.d_close =	nf2_close,
	.d_name =	"netfpga",
};

/*----------------------------------------------------------------------------
 * Helper functions.
 */

static void __unused
_nf2_resource_debug(const char *msg, struct resource *r)
{
	u_long rstart, rend, rsz;
	void *va;
	int rid;

	NF2_ASSERT(msg != NULL);
	NF2_ASSERT(r != NULL);

	rstart = rman_get_start(r);
	rend = rman_get_end(r);
	rsz = rman_get_size(r);
	va = rman_get_virtual(r);
	rid = rman_get_rid(r);

	printf("<Resource: %s>\n", msg);
	printf("Region: %#lx-%#lx (size: %#lx [%ldMiB])\n", rstart, rend,
	    rsz, rsz / (1024 * 1024));
	printf("Address: %p\n", va);
	printf("   RID: %d\n", rid);
}

/*--------------------------------------------------------------------------*/

static void
nf2_write_flush(struct nf2_softc *sc)
{

	NF2_SOFTC_ASSERT(sc);
	RD32(sc, 0);
}

static void
nf2_mac_reset(struct nf2_softc *sc, int num)
{
	uint32_t mac;

	NF2_SOFTC_ASSERT(sc);

	/* Reset particular MAC */
	mac = CNET_RESET_MAC_0 << num;
	WR32(sc, CNET_REG_RESET, mac);
	nf2_write_flush(sc);
}

static void
nf2_mac_reset_all(struct nf2_softc *sc)
{
	int i;

	NF2_SOFTC_ASSERT(sc);
	for (i = 0; i < MAX_IFACE; i++)
		nf2_mac_reset(sc, i);
}

static void
nf2_mac_enable(struct nf2_softc *sc, int num)
{
	uint32_t enable;

	NF2_SOFTC_ASSERT(sc);

	/* Read the old register value */
	enable = RD32(sc, CNET_REG_ENABLE);
	NF2_DEBUG("enable=%#x (%d)", enable, enable);

	/* Enable particular MAC */
	enable |= CNET_ENABLE_RX_FIFO_0 << num;
	enable |= CNET_ENABLE_TX_MAC_0 << num;
	enable |= CNET_ENABLE_INGRESS_ARBITER;
	enable |= CNET_ENABLE_RX_DMA;

	/* Write updated value to the ENABLE register */
	NF2_DEBUG("enable set to %#x (%d)", enable, enable);
	WR32(sc, CNET_REG_ENABLE, enable);

	/* See if the value actually got changed */
	enable = RD32(sc, CNET_REG_ENABLE);
	NF2_DEBUG("enable read is %#x (%d)", enable, enable);

	nf2_write_flush(sc);
}

static void
nf2_mac_enable_all(struct nf2_softc *sc)
{
	int i;

	NF2_SOFTC_ASSERT(sc);
	for (i = 0; i < MAX_IFACE; i++)
		nf2_mac_enable(sc, i);
}

static void
nf2_mac_disable(struct nf2_softc *sc, int num)
{
	uint32_t enable;

	NF2_SOFTC_ASSERT(sc);

	NF2_DEBUG("--");
	/* Read the old register value */
	enable = RD32(sc, CNET_REG_ENABLE);
	NF2_DEBUG("Denable=%#x (%d)", enable, enable);

	/* Disable particular MAC */
	enable &= ~CNET_ENABLE_RX_FIFO_0 << num;
	enable &= ~CNET_ENABLE_TX_MAC_0 << num;
#if 0
	/*
	 * XXWKOSZEK:
	 * See, why those flags are missing in a Linux kernel driver.
	 * They get set when MAC is being enabled, but don't get
	 * cleared.
	 */
	enable &= ~CNET_ENABLE_INGRESS_ARBITER;
	enable &= ~CNET_ENABLE_RX_DMA;
#endif

	/* Write updated value to the ENABLE register */
	NF2_DEBUG("Denable set to %#x (%d)", enable, enable);
	WR32(sc, CNET_REG_ENABLE, enable);

	/* See if the value actually got changed */
	enable = RD32(sc, CNET_REG_ENABLE);
	NF2_DEBUG("Denable read is %#x (%d)", enable, enable);

	nf2_write_flush(sc);
}

static void
nf2_mac_disable_all(struct nf2_softc *sc)
{
	int i;

	NF2_SOFTC_ASSERT(sc);
	for (i = 0; i < MAX_IFACE; i++)
		nf2_mac_disable(sc, i);

	/* 
	 * XXWKOSZEK: Map the Linux's code for now.
	 * See nf2_mac_disable for explanation.
	 */
	WR32(sc, CNET_REG_ENABLE, 0);
}

static void
nf2_cpci_reset(struct nf2_softc *sc)
{

	NF2_SOFTC_ASSERT(sc);
	WR32(sc, CPCI_REG_RESET, RESET_CPCI);
	nf2_write_flush(sc);
}

static void
nf2_irq_enable(struct nf2_softc *sc)
{

	NF2_SOFTC_ASSERT(sc);
	/* Enable interrupts */
	WR32(sc, CPCI_REG_INTERRUPT_MASK, 0);
	nf2_write_flush(sc);
}

static void __unused
nf2_irq_disable(struct nf2_softc *sc)
{

	NF2_SOFTC_ASSERT(sc);
	/* Disable interrupts */
	WR32(sc, CPCI_REG_INTERRUPT_MASK, ~0);
	nf2_write_flush(sc);
}

static void
nf2_setup(struct nf2_softc *sc)
{
	uint32_t macs;

	NF2_SOFTC_ASSERT(sc);

	/* Reset the CPCI */
	nf2_cpci_reset(sc);

	/* 
	 * XXWKOSZEK:
	 * Reset CNET. This is direct port of the Linux code. It doesn't
	 * seem to make much sense to reset MACs in here and repeat it
	 * later, so I belive we should be sending only CTRL_CNET_RESET
	 * here and reseting MACs via nf2_mac_reset_all() next.
	 */
	macs =
	    CNET_RESET_MAC_3,
	    CNET_RESET_MAC_2,
	    CNET_RESET_MAC_1,
	    CNET_RESET_MAC_0;
	WR32(sc, CNET_REG_CTRL, CTRL_CNET_RESET);
	WR32(sc, CNET_REG_RESET, macs);

	/* Flush the writes */
	nf2_write_flush(sc);

	/* Disable all MACs */
	nf2_mac_disable_all(sc);

	/* Reset all MACs */
	nf2_mac_reset_all(sc);

	/* Disable interrupts */
	nf2_irq_disable(sc);
}

static void
nf2_go(struct nf2_softc *sc)
{

	NF2_SOFTC_ASSERT(sc);

	/* Enable all MACs */
	nf2_mac_enable_all(sc);

	/* Enable interrupts */
	nf2_irq_enable(sc);
}

static void __unused
nf2_intr(void *arg)
{
	struct nf2_softc *sc;
	uint32_t mask;
	uint32_t status;

	sc = arg;
	NF2_SOFTC_ASSERT(sc);
#if 0
	nf2_irq_disable(sc);
#endif
	mask = RD32(sc, CPCI_REG_INTERRUPT_MASK);
	status = RD32(sc, CPCI_REG_INTERRUPT_STATUS);
	status &= ~(mask & ~INT_PKT_AVAIL);
#define PRINT_IRQ(x)						\
	if (status & (x) && printf("%s, status=%x", (#x), (status)))
	PRINT_IRQ(INT_DMA_RX_COMPLETE) {}
	PRINT_IRQ(INT_DMA_TX_COMPLETE) {}
	PRINT_IRQ(INT_PHY_INTERRUPT) {}
	PRINT_IRQ(INT_PKT_AVAIL) {
		mask |= ~INT_PKT_AVAIL;
	}
	PRINT_IRQ(INT_CNET_ERROR) {}
	PRINT_IRQ(INT_CNET_READ_TIMEOUT) {}
	PRINT_IRQ(INT_PROG_ERROR) {}
	PRINT_IRQ(INT_DMA_TRANSFER_ERROR) {}
	PRINT_IRQ(INT_DMA_SETUP_ERROR) {}
	PRINT_IRQ(INT_DMA_FATAL_ERROR) {}
#if 0
	WR32(sc, CPCI_REG_INTERRUPT_MASK, mask);
	nf2_irq_enable(sc);
#endif
}

static int
nf2_dma_alloc(struct nf2_softc *sc)
{
	int error = 0;

	NF2_SOFTC_ASSERT(sc);

	/*
	 * Allocate the parent bus DMA tag appropriate for PCI.
	 */
	error = bus_dma_tag_create(
	    bus_get_dma_tag(sc->dev),	/* parent */
	    1,				/* alignment */
	    0,				/* boundary */
	    BUS_SPACE_MAXADDR_32BIT,	/* lowaddr */
	    BUS_SPACE_MAXADDR,		/* highaddr */
	    NULL,			/* filter */
	    NULL,			/* filterarg */
	    BUS_SPACE_MAXSIZE_32BIT,	/* maxsize */
	    0,				/* nsegments */
	    BUS_SPACE_MAXSIZE_32BIT,	/* maxsegsize */
	    0,				/* flags */
	    NULL,			/* lockfunc */
	    NULL,			/* lockarg */
	    &sc->parent_tag);
	if (error) {
		NF2_DEBUG("Couldn't create parent DMA tag: %d", error);
		return (error);
	}

	/* Create DMA tag for Rx memory. */
	error = bus_dma_tag_create(
	    sc->parent_tag,		/* parent */
	    1,				/* alignment */
	    0,				/* boundary */
	    BUS_SPACE_MAXADDR_32BIT,	/* lowaddr */
	    BUS_SPACE_MAXADDR,		/* highaddr */
	    NULL,			/* filter */
	    NULL,			/* filterarg */
	    BUS_SPACE_MAXSIZE_32BIT,	/* maxsize */
	    0,				/* nsegments */
	    BUS_SPACE_MAXSIZE_32BIT,	/* maxsegsize */
	    0,				/* flags */
	    NULL,			/* lockfunc */
	    NULL,			/* lockarg */
	    &sc->rx_tag);
	if (error) {
		NF2_DEBUG("Couldn't create DMA tag for RX mem: %d", error);
		return (error);
	}
	return (error);
}

static int
nf2_dma_free(struct nf2_softc *sc)
{
	int error = 0;

	NF2_SOFTC_ASSERT(sc);
	if (sc->parent_tag != NULL)
		error = bus_dma_tag_destroy(sc->parent_tag);
	if (sc->rx_tag != NULL)
		error = bus_dma_tag_destroy(sc->rx_tag);
	NF2_DEBUG("returning with error=%d", error);
	return (error);
}

static void
nf2_test(struct nf2_softc *sc)
{
	uint32_t r;
	char msg[100];
	uint32_t *msgptr;

#define P(x) do {					\
	r = RD32(sc, (x));				\
	NF2_DEBUG("%s=%#x (%d)\n", #x, (r), (r));	\
} while (0)

	P(NF2_DEVICE_ID);
	P(NF2_REVISION);
#define DEVICE_STR_LEN 120
	int i;
	msgptr = (uint32_t *)msg;
	for (i = 0; i < DEVICE_STR_LEN / 4; i += 4) {
		*msgptr = htonl(RD32(sc, 0x0400008 + i));
		msgptr++;
	}

	NF2_DEBUG("String read: '%s'\n", msg);
#if 0
	P(CPCI_REG_DMA_I_SIZE);
	DELAY(1000000);

	P(CNET_REG_ENABLE);
	DELAY(1000000);

	WR32(sc, CNET_REG_ENABLE, 0);
	DELAY(1000000);

	P(CNET_REG_ENABLE);
	DELAY(1000000);
#endif
}

/*--------------------------------------------------------------------------*/

static int
nf2_probe(device_t dev)
{
	uint16_t vid;
	uint16_t did;

	vid = pci_get_vendor(dev);
	did = pci_get_device(dev);

	if (vid == VENDOR_ID_STANFORD && did == DEVICE_ID_STANFORD_NF2) {
		device_set_desc(dev, "NetFPGA rev 2.");
		NF2_DEBUG("Probe succeed (dev=%p, VID: %#x, DID: %#x)", dev, vid, did);
		return (BUS_PROBE_DEFAULT);
	}
	return (ENXIO);
}

static int
nf2_attach(device_t dev)
{
	struct nf2_softc *sc = NULL;
	int error = 0;
	int rid = 0;
	int unit = 0;

	(void)rid;

	NF2_DEBUG("Attach called (dev=%p)", dev);
	/*
	 * Get the whole device's context right.
	 */
	sc = device_get_softc(dev);
	NF2_SOFTC_INIT(sc);
	NF2_SOFTC_ASSERT(sc);
	sc->mem = NULL;
	sc->ioport = NULL;
	sc->irq = NULL;
	sc->dev = dev;

	/*
	 * Enable bus mastering.
	 */
	pci_enable_busmaster(dev);

	sc->revid = pci_get_revid(dev);
	if (sc->revid != 0) {
		NF2_DEBUG("Wrong card's revision: %d (expected)", sc->revid);
		return (ENXIO);
	}

	/* Allocate a memory region */
	rid = PCIR_BAR(0);
	sc->mem = bus_alloc_resource_any(dev, SYS_RES_MEMORY, &rid, RF_ACTIVE);
	if (sc->mem == NULL) {
		NF2_DEBUG("Couldn't allocate memory!\n");
		error = ENXIO;
		goto errout;
	}
	sc->mem_tag = rman_get_bustag(sc->mem);
	sc->mem_handle = rman_get_bushandle(sc->mem);

#if 0
	/* 
	 * Allocate a I/O space
	 * See why it's 
	 */
	rid = PCIR_BAR(2);
	sc->ioport = bus_alloc_resource_any(dev, SYS_RES_IOPORT, &rid, RF_ACTIVE);
	if (sc->ioport == NULL) {
		NF2_DEBUG("Couldn't allocate I/O space!\n");
		error = ENXIO;
		goto errout;
	}
#endif

	/* Allocate interrupt. */
	rid = 0;
	sc->irq = bus_alloc_resource_any(dev, SYS_RES_IRQ, &rid,
	    RF_SHAREABLE | RF_ACTIVE);
	if (sc->irq == NULL) {
		NF2_DEBUG("Couldn't allocate interrupt");
		error = ENXIO;
		goto errout;
	}

	if (sc->mem != NULL)
		_nf2_resource_debug("memory", sc->mem);
	if (sc->ioport != NULL)
		_nf2_resource_debug("I/O space", sc->ioport);
	if (sc->irq != NULL)
		_nf2_resource_debug("irq", sc->irq);

	sc->need_reset = 0;

	/*
	 * /dev/netfpga<NUM> registration.
	 */
	unit = device_get_unit(dev);
	sc->cdev = make_dev(&nf2_cdevsw, unit, UID_ROOT, GID_OPERATOR,
	    0600, "netfpga%d", unit);
	NF2_ASSERT(sc->cdev != NULL);
	sc->cdev->si_drv1 = sc;

if (0)
	nf2_setup(sc);
if (0)
	nf2_dma_alloc(sc);
	error = bus_setup_intr(dev, sc->irq, INTR_TYPE_NET | INTR_MPSAFE,
	    NULL, nf2_intr, sc, &sc->intrhand);
	if (error != 0) {
		NF2_DEBUG("Couldn't setup an interrupt");
		goto errout;
	}

	nf2_test(sc);
if (0)
	nf2_go(sc);

errout:
	if (error != 0) {
		NF2_DEBUG("Returning with error code %d\n", error);
		error = nf2_detach(dev);
	}
	return (error);
}

static int
nf2_detach(device_t dev)
{
	struct nf2_softc *sc = NULL;
	int error = 0;
	int rid = 0;

	(void)error;
	(void)rid;

	sc = device_get_softc(dev);
	NF2_SOFTC_ASSERT(sc);

	if (sc->intrhand != NULL) {
		error = bus_teardown_intr(dev, sc->irq, sc->intrhand);
		if (error != 0)
			NF2_DEBUG("Couldn't tear down interrupt");
	}

	nf2_dma_free(sc);

	if (sc->cdev)
		destroy_dev(sc->cdev);

	if (sc->mem != NULL) {
		rid = rman_get_rid(sc->mem);
		error = bus_release_resource(dev, SYS_RES_MEMORY, rid, sc->mem);
		if (error != 0)
			NF2_DEBUG("Couldn't release memory!");
	}

	if (sc->ioport != NULL) {
		rid = rman_get_rid(sc->ioport);
		error = bus_release_resource(dev, SYS_RES_IOPORT, rid, sc->ioport);
		if (error != 0)
			NF2_DEBUG("Couldn't release I/O space!");
	}

	if (sc->irq != NULL) {
		rid = rman_get_rid(sc->irq);
		error = bus_release_resource(dev, SYS_RES_IRQ, rid, sc->irq);
		if (error != 0)
			NF2_DEBUG("Couldn't release IRQ!");
	}

	return (error);
}

/*--------------------------------------------------------------------------*/

static int
nf2_open(struct cdev *dev, int flags, int fmp, struct thread *td)
{

	NF2_DEBUG("Open called");
	return (0);
}

static int
nf2_close(struct cdev *dev, int flags, int fmt, struct thread *td)
{
	struct nf2_softc *sc = NULL;

	sc = dev->si_drv1;
	NF2_SOFTC_ASSERT(sc);
	NF2_DEBUG("Close called");

	if (sc->need_reset != 0) {
		NF2_DEBUG("Reset of device needed (code=%d)! Performing..", sc->need_reset);
		nf2_mac_disable_all(sc);
		nf2_mac_reset_all(sc);
		nf2_cpci_reset(sc);
		NF2_DEBUG("Reset done");
	}

	return (0);
}

static int
nf2_ioctl(struct cdev *dev, unsigned long cmd, caddr_t data, int fflag,
    struct thread *td)
{
	struct nf2_req *req = NULL;
	struct nf2_softc *sc = NULL;
	int error = 0;

	sc = dev->si_drv1;
	NF2_SOFTC_ASSERT(sc);
	NF2_DEBUG3("NF2 ioctl() called with sc=%p", sc);
	req = (struct nf2_req *)data;
	NF2_DEBUG3("NF2 ioctl() req.offset=%#llx, req.value=%#llx",
	    req->offset, req->value);

	/* CPCI reprogramming */
	if (req->offset >= VIRTEX_PROGRAM_RAM_BASE_ADDR &&
	    req->offset <= VIRTEX_PROGRAM_RAM_BASE_ADDR + CPCI_BIN_SIZE)
		sc->need_reset = 1;
	/* Virtex programming */
	if (req->offset == CPCI_PROGRAMMING_DATA)
		sc->need_reset = 2;

	switch (cmd) {
	case SIOCREGREAD:
		NF2_DEBUG3("SIOCREGREAD");
		req->value = RD32(sc, req->offset);
		break;
	case SIOCREGWRITE:
		NF2_DEBUG3("SIOCREGWRITE");
		WR32(sc, req->offset, req->value);
		break;
	default:
		error = EINVAL;
	}
	NF2_DEBUG3("ioctl() error = 0");
	return (error);
}
